var Racer, express, fs, gzip;
express = require('express');
gzip = require('connect-gzip');
fs = require('fs');
Racer = require('racer').Racer;
module.exports = function(racer) {
  var app, initRoom, store;
  exports.app = app = express.createServer().use(express.favicon()).use('/letters', gzip.staticGzip(__dirname));
  if (!racer) {
    racer = new Racer({
      redis: {
        db: 1
      },
      listen: app
    });
  }
  store = racer.store;
  store.flush();
  racer.js({
    entry: __dirname + '/client.js'
  }, function(js) {
    return fs.writeFileSync(__dirname + '/script.js', js);
  });
  app.get('/letters/:room?', function(req, res) {
    var room, _room;
    room = req.params.room;
    if (!(room && /^[-\w ]+$/.test(room))) {
      return res.redirect('/letters/lobby');
    }
    _room = room.toLowerCase().replace(/[_ ]/g, '-');
    if (_room !== room) {
      return res.redirect("/letters/" + _room);
    }
    return store.subscribe({
      _room: "rooms." + room + ".**"
    }, 'rooms.*.players', function(err, model) {
      model.set('_roomName', room);
      initRoom(model);
      return model.bundle(function(bundle) {
        var boardHtml, close, id, letter, open, _ref;
        boardHtml = '';
        if (~req.headers['user-agent'].indexOf('MSIE')) {
          open = '<a href=# onclick="return false"';
          close = '</a>';
        } else {
          open = '<span';
          close = '</span>';
        }
        _ref = model.get("_room.letters");
        for (id in _ref) {
          letter = _ref[id];
          boardHtml += "" + open + " draggable=true class=\"" + letter.color + " letter\" id=" + id + "\nstyle=left:" + letter.position.left + "px;top:" + letter.position.top + "px>" + letter.value + close;
        }
        return res.send("<!DOCTYPE html>\n<title>Letters game</title>\n<link rel=stylesheet href=style.css>\n<link rel=stylesheet href=http://fonts.googleapis.com/css?family=Anton>\n<div id=back>\n  <div id=page>\n    <p id=info>\n    <div id=rooms>\n      <p>Rooms:\n      <ul id=roomlist></ul>\n    </div>\n    <div id=board>" + boardHtml + "</div>\n  </div>\n</div>\n<script>init=" + bundle + "</script>\n<script src=script.js></script>");
      });
    });
  });
  initRoom = function(model) {
    var col, colors, letters, row;
    if (model.get('_room.letters')) {
      return;
    }
    colors = ['red', 'yellow', 'blue', 'orange', 'green'];
    letters = {};
    for (row = 0; row <= 4; row++) {
      for (col = 0; col <= 25; col++) {
        letters[row * 26 + col] = {
          color: colors[row],
          value: String.fromCharCode(65 + col),
          position: {
            left: col * 24 + 72,
            top: row * 32 + 8
          }
        };
      }
    }
    return model.set('_room.letters', letters);
  };
  racer.sockets.on('connection', function(socket) {
    return socket.on('join', function(room) {
      var playersPath;
      playersPath = "rooms." + room + ".players";
      store.incr(playersPath);
      return socket.on('disconnect', function() {
        return store.incr(playersPath, -1);
      });
    });
  });
  return exports;
};