var EventEmitter, MemorySync, Model, RESEND_INTERVAL, RefHelper, SEND_TIMEOUT, TxnApplier, arrayMethod, merge, pathParser, specHelper, transaction;
var __slice = Array.prototype.slice;
transaction = require('./transaction');
pathParser = require('./pathParser');
MemorySync = require('./adapters/MemorySync');
TxnApplier = require('./TxnApplier');
RefHelper = require('./RefHelper');
specHelper = require('./specHelper');
EventEmitter = require('events').EventEmitter;
merge = require('./util').merge;
Model = module.exports = function(_clientId, AdapterClass) {
  var adapter, refHelper, self, txnApplier, txnQueue, txns;
  this._clientId = _clientId != null ? _clientId : '';
  if (AdapterClass == null) {
    AdapterClass = MemorySync;
  }
  self = this;
  self._adapter = adapter = new AdapterClass;
  self._initRefs(adapter);
  self._cache = {
    invalidateSpecModelCache: function() {
      delete this.obj;
      delete this.lastReplayedTxnId;
      return delete this.path;
    }
  };
  self._storeSubs = [];
  self._startId = '';
  self._txnCount = 0;
  self._txns = txns = {};
  self._txnQueue = txnQueue = [];
  txnApplier = new TxnApplier({
    applyTxn: function(txn) {
      if (transaction.base(txn) > adapter.ver) {
        return self._applyTxn(txn, !txn.emitted && this._clientId !== transaction.clientId(txn));
      }
    },
    onTimeout: function() {
      return self._reqNewTxns();
    }
  });
  self._onTxn = function(txn, num) {
    var queuedTxn;
    if (queuedTxn = txns[transaction.id(txn)]) {
      txn.callback = queuedTxn.callback;
      txn.emitted = queuedTxn.emitted;
    }
    return txnApplier.add(txn, num);
  };
  self._onTxnNum = function(num) {
    txnApplier.setIndex((+num || 0) + 1);
    return txnApplier.clearPending();
  };
  self._removeTxn = function(txnId) {
    var i;
    delete txns[txnId];
    if (~(i = txnQueue.indexOf(txnId))) {
      txnQueue.splice(i, 1);
    }
    return self._cache.invalidateSpecModelCache();
  };
  self.force = Object.create(self, {
    _force: {
      value: true
    }
  });
  refHelper = self._refHelper;
  ['set', 'del', 'push', 'pop', 'insertAfter', 'insertBefore', 'remove', 'splice', 'move'].forEach(function(method) {
    return self.on(method, function(_arg) {
      var args, path;
      path = _arg[0], args = 2 <= _arg.length ? __slice.call(_arg, 1) : [];
      return refHelper.notifyPointersTo(path, this.get(), method, args);
    });
  });
};
Model.prototype = {
  _commit: function() {},
  _reqNewTxns: function() {},
  _setSocket: function(socket) {
    var adapter, clientId, commit, connection, fatalErr, onConnectionStatus, onTxn, removeTxn, resend, resendInterval, self, storeSubs, txnQueue, txns;
    this.socket = socket;
    self = this;
    adapter = this._adapter;
    txns = this._txns;
    txnQueue = this._txnQueue;
    onTxn = this._onTxn;
    removeTxn = this._removeTxn;
    this._commit = commit = function(txn) {
      if (!socket.socket.connected) {
        return;
      }
      txn.timeout = +new Date + SEND_TIMEOUT;
      return socket.emit('txn', txn, self._startId);
    };
    this._reqNewTxns = function() {
      return socket.emit('txnsSince', adapter.ver + 1, self._startId);
    };
    socket.on('txn', onTxn);
    socket.on('txnNum', this._onTxnNum);
    socket.on('txnOk', function(txnId, base, num) {
      var txn;
      if (txn = txns[txnId]) {
        transaction.base(txn, base);
        return onTxn(txn, num);
      }
    });
    socket.on('txnErr', function(err, txnId) {
      var args, callback, txn;
      txn = txns[txnId];
      if (txn && (callback = txn.callback)) {
        args = transaction.args(txn);
        args.unshift(err);
        callback.apply(null, args);
      }
      return removeTxn(txnId);
    });
    fatalErr = false;
    socket.on('fatalErr', function() {
      fatalErr = true;
      return socket.disconnect();
    });
    this.connection = connection = function() {
      if (fatalErr) {
        return null;
      }
      return socket.socket.connected;
    };
    onConnectionStatus = function() {
      return self.emit('connectionStatus', connection());
    };
    clientId = this._clientId;
    storeSubs = this._storeSubs;
    resendInterval = null;
    resend = function() {
      var id, now, txn, _i, _len, _results;
      now = +new Date;
      _results = [];
      for (_i = 0, _len = txnQueue.length; _i < _len; _i++) {
        id = txnQueue[_i];
        txn = txns[id];
        if (txn.timeout > now) {
          return;
        }
        _results.push(commit(txn));
      }
      return _results;
    };
    socket.on('connect', function() {
      var id, _i, _len;
      onConnectionStatus();
      socket.emit('sub', clientId, storeSubs, adapter.ver, self._startId);
      for (_i = 0, _len = txnQueue.length; _i < _len; _i++) {
        id = txnQueue[_i];
        commit(txns[id]);
      }
      if (!resendInterval) {
        return resendInterval = setInterval(resend, RESEND_INTERVAL);
      }
    });
    socket.on('disconnect', function() {
      if (resendInterval) {
        clearInterval(resendInterval);
      }
      resendInterval = null;
      return setTimeout(onConnectionStatus, 200);
    });
    return socket.on('connect_failed', onConnectionStatus);
  },
  _nextTxnId: function() {
    return this._clientId + '.' + this._txnCount++;
  },
  _addTxn: function() {
    var $k, $r, args, callback, firstArgs, id, idAsIndex, members, meta, method, model, path, refHelper, txn, txnArgs, ver, _i, _ref, _ref2;
    method = arguments[0], path = arguments[1], args = 4 <= arguments.length ? __slice.call(arguments, 2, _i = arguments.length - 1) : (_i = 2, []), callback = arguments[_i++];
    refHelper = this._refHelper;
    model = this;
    if (_ref = refHelper.isArrayRef(path, this._specModel()[0]), $r = _ref.$r, $k = _ref.$k, _ref) {
      _ref2 = refHelper.splitArrayArgs(method, args), firstArgs = _ref2[0], members = _ref2[1];
      members = members.map(function(member) {
        if (refHelper.isRefObj(member)) {
          return member;
        }
        model.set($r + '.' + member.id, member);
        return {
          $r: $r,
          $k: member.id.toString()
        };
      });
      args = firstArgs.concat(members);
    }
    if (arrayMethod[method]) {
      idAsIndex = refHelper.arrRefIndex(args[0], path, this._specModel()[0]);
    }
    ver = this._force ? null : this._adapter.ver;
    id = this._nextTxnId();
    txn = transaction.create({
      base: ver,
      id: id,
      method: method,
      args: [path].concat(__slice.call(args))
    });
    txn = refHelper.dereferenceTxn(txn, this._specModel()[0]);
    this._txns[id] = txn;
    txn.callback = callback;
    this._txnQueue.push(id);
    txnArgs = transaction.args(txn);
    path = txnArgs[0];
    if (pathParser.isPrivate(path)) {
      this._cache.invalidateSpecModelCache();
      return this._applyTxn(txn, !txn.emitted && true);
    }
    if (idAsIndex !== void 0) {
      meta = txnArgs[1];
      meta.index = idAsIndex;
      transaction.meta(txn, meta);
    }
    this.emit(method, txnArgs);
    txn.emitted = true;
    if (idAsIndex !== void 0) {
      txnArgs[1] = idAsIndex;
    }
    return this._commit(txn);
  },
  _applyTxn: function(txn, forceEmit) {
    var args, callback, meta, method, txnArgs, _ref;
    method = transaction.method(txn);
    txnArgs = transaction.args(txn);
    args = txnArgs.slice(0);
    args.push(transaction.base(txn));
    (_ref = this._adapter)[method].apply(_ref, args);
    this._removeTxn(transaction.id(txn));
    if (forceEmit) {
      if (meta = transaction.meta(txn)) {
        args[1] = meta;
      }
      this.emit(method, args);
    }
    if (callback = txn.callback) {
      return callback.apply(null, [null].concat(__slice.call(txnArgs)));
    }
  },
  _specModel: function() {
    var adapter, args, cache, i, lastReplayedTxnId, len, meta, obj, path, replayFrom, txn;
    cache = this._cache;
    if (lastReplayedTxnId = cache.lastReplayedTxnId) {
      if (cache.lastReplayedTxnId === this._txnQueue[this._txnQueue.length - 1]) {
        return [cache.obj, cache.path];
      }
      obj = cache.obj;
      replayFrom = 1 + this._txnQueue.indexOf(cache.lastReplayedTxnId);
    } else {
      replayFrom = 0;
    }
    adapter = this._adapter;
    if (len = this._txnQueue.length) {
      if (!obj) {
        obj = cache.obj = specHelper.create(adapter._data);
      }
      i = replayFrom;
      while (i < len) {
        txn = this._txns[this._txnQueue[i++]];
        args = transaction.args(txn).slice(0);
        args.push(adapter.ver, {
          obj: obj,
          proto: true,
          returnMeta: true
        });
        meta = adapter[transaction.method(txn)].apply(adapter, args);
        path = meta.path;
      }
      cache.obj = obj;
      cache.path = path;
      cache.lastReplayedTxnId = transaction.id(txn);
    }
    return [obj, path];
  },
  _initRefs: function(adapter) {
    var refHelper;
    return this._refHelper = refHelper = new RefHelper(this);
  },
  ref: RefHelper.prototype.ref,
  arrayRef: RefHelper.prototype.arrayRef,
  get: function(path) {
    return this._adapter.get(path, this._specModel()[0]);
  },
  set: function(path, value, callback) {
    this._addTxn('set', path, value, callback);
    return value;
  },
  del: function(path, callback) {
    return this._addTxn('del', path, callback);
  },
  incr: function(path, byNum, callback) {
    if (typeof byNum === 'function') {
      callback = byNum;
      byNum = 1;
    } else if (typeof byNum !== 'number') {
      byNum = 1;
    }
    return this.set(path, (this.get(path) || 0) + byNum, callback);
  },
  push: function() {
    var callback, path, values, _i;
    path = arguments[0], values = 3 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 1) : (_i = 1, []), callback = arguments[_i++];
    if ('function' !== typeof callback && callback !== void 0) {
      values.push(callback);
      callback = null;
    }
    return this._addTxn.apply(this, ['push', path].concat(__slice.call(values), [callback]));
  },
  pop: function(path, callback) {
    return this._addTxn('pop', path, callback);
  },
  unshift: function() {
    var callback, path, values, _i;
    path = arguments[0], values = 3 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 1) : (_i = 1, []), callback = arguments[_i++];
    if ('function' !== typeof callback && callback !== void 0) {
      values.push(callback);
      callback = null;
    }
    return this._addTxn.apply(this, ['unshift', path].concat(__slice.call(values), [callback]));
  },
  shift: function(path, callback) {
    return this._addTxn('shift', path, callback);
  },
  insertAfter: function(path, afterIndex, value, callback) {
    return this._addTxn('insertAfter', path, afterIndex, value, callback);
  },
  insertBefore: function(path, beforeIndex, value, callback) {
    return this._addTxn('insertBefore', path, beforeIndex, value, callback);
  },
  remove: function(path, start, howMany, callback) {
    if (howMany == null) {
      howMany = 1;
    }
    if (typeof howMany === 'function') {
      callback = howMany;
      howMany = 1;
    }
    return this._addTxn('remove', path, start, howMany, callback);
  },
  splice: function() {
    var callback, newMembers, path, removeCount, startIndex, _i;
    path = arguments[0], startIndex = arguments[1], removeCount = arguments[2], newMembers = 5 <= arguments.length ? __slice.call(arguments, 3, _i = arguments.length - 1) : (_i = 3, []), callback = arguments[_i++];
    if ('function' !== typeof callback && callback !== void 0) {
      newMembers.push(callback);
      callback = null;
    }
    return this._addTxn.apply(this, ['splice', path, startIndex, removeCount].concat(__slice.call(newMembers), [callback]));
  },
  move: function(path, from, to, callback) {
    return this._addTxn('move', path, from, to, callback);
  }
};
Model._SEND_TIMEOUT = SEND_TIMEOUT = 10000;
Model._RESEND_INTERVAL = RESEND_INTERVAL = 2000;
arrayMethod = {
  remove: 1,
  insertAfter: 1,
  insertBefore: 1,
  splice: 1,
  move: 1
};
merge(Model.prototype, EventEmitter.prototype);
Model.prototype._eventListener = function(method, pattern, callback) {
  var re;
  if (pattern.call) {
    return pattern;
  }
  re = pathParser.regExp(pattern);
  return function(_arg) {
    var args, path;
    path = _arg[0], args = 2 <= _arg.length ? __slice.call(_arg, 1) : [];
    if (re.test(path)) {
      return callback.apply(null, re.exec(path).slice(1).concat(args));
    }
  };
};
Model.prototype._on = EventEmitter.prototype.on;
Model.prototype.on = Model.prototype.addListener = function(type, pattern, callback) {
  var listener;
  this._on(type, listener = this._eventListener(type, pattern, callback));
  return listener;
};
Model.prototype._once = EventEmitter.prototype.once;
Model.prototype.once = function(type, pattern, callback) {
  var listener;
  this._once(type, listener = this._eventListener(type, pattern, callback));
  return listener;
};