var ARRAY_OPS, RefHelper, argsNormalizer, hasKeys, merge, pathParser, specHelper, transaction, _ref;
var __slice = Array.prototype.slice, __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
transaction = require('./transaction');
pathParser = require('./pathParser');
specHelper = require('./specHelper');
_ref = require('./util'), merge = _ref.merge, hasKeys = _ref.hasKeys;
argsNormalizer = require('./argsNormalizer');
module.exports = RefHelper = function(model) {
  this._model = model;
  this._adapter = model._adapter;
  this._setup();
};
ARRAY_OPS = {
  push: 1,
  unshift: 1,
  pop: 1,
  shift: 1,
  remove: 1,
  insertAfter: 1,
  insertBefore: 1,
  splice: 1,
  move: 1
};
RefHelper.prototype = {
  isPointer: function(val) {
    return !!(val && val.$r);
  },
  isKeyPath: function(path, obj) {
    var found;
    if (!obj) {
      throw new Error('Missing obj');
    }
    found = this._adapter.lookup("$keys." + path + ".$", false, {
      obj: obj
    }).obj;
    return found !== void 0;
  },
  isPathPointedTo: function(path, obj) {
    var found;
    if (!obj) {
      throw 'Missing obj';
    }
    found = this._adapter.lookup("$refs." + path + ".$", false, {
      obj: obj
    }).obj;
    return found !== void 0;
  },
  _setup: function() {
    var adapter, refHelper;
    refHelper = this;
    adapter = this._adapter;
    adapter.__set = adapter.set;
    adapter.set = function(path, value, ver, options) {
      var out;
      if (options == null) {
        options = {};
      }
      options.obj || (options.obj = this._data);
      if (refHelper.isPointer(value)) {
        refHelper.$indexRefs(path, value.$r, value.$k, value.$t, ver, options);
      }
      out = this.__set(path, value, ver, options);
      refHelper.updateRefsForKey(path, ver, options);
      return out;
    };
    adapter.__del = adapter.del;
    adapter.del = function(path, ver, options) {
      var out;
      if (options == null) {
        options = {};
      }
      options.obj || (options.obj = this._data);
      out = this.__del(path, ver, options);
      if (refHelper.isPathPointedTo(path, options.obj)) {
        refHelper.cleanupPointersTo(path, options);
      }
      return out;
    };
    adapter.__remove = adapter.remove;
    adapter.remove = function(path, start, howMany, ver, options) {
      var out, startIndex;
      if (options == null) {
        options = {};
      }
      options.obj || (options.obj = this._data);
      startIndex = refHelper.arrRefIndex(start, path, options.obj);
      out = this.__remove(path, startIndex, howMany, ver, options);
      refHelper.updateRefsForKey(path, ver, options);
      return out;
    };
    adapter.__move = adapter.move;
    adapter.move = function(path, from, to, ver, options) {
      var out, startIndex;
      if (options == null) {
        options = {};
      }
      options.obj || (options.obj = this._data);
      startIndex = refHelper.arrRefIndex(from, path, options.obj);
      out = this.__move(path, startIndex, to, ver, options);
      refHelper.updateRefsForKey(path, ver, options);
      return out;
    };
    ['push', 'unshift'].forEach(function(method) {
      adapter['__' + method] = adapter[method];
      return adapter[method] = function() {
        var options, out, path, values, ver, _ref2;
        _ref2 = argsNormalizer.adapter[method](arguments), path = _ref2[0], values = _ref2[1], ver = _ref2[2], options = _ref2[3];
        options.obj || (options.obj = this._data);
        out = this['__' + method].apply(this, [path].concat(__slice.call(values), [ver], [options]));
        refHelper.updateRefsForKey(path, ver, options);
        return out;
      };
    });
    ['pop', 'shift'].forEach(function(method) {
      adapter['__' + method] = adapter[method];
      return adapter[method] = function(path, ver, options) {
        var out;
        if (options == null) {
          options = {};
        }
        options.obj || (options.obj = this._data);
        out = this['__' + method](path, ver, options);
        refHelper.updateRefsForKey(path, ver, options);
        return out;
      };
    });
    ['insertAfter', 'insertBefore'].forEach(function(method) {
      adapter['__' + method] = adapter[method];
      return adapter[method] = function(path, pos, value, ver, options) {
        var index, out;
        if (options == null) {
          options = {};
        }
        options.obj || (options.obj = this._data);
        index = refHelper.arrRefIndex(pos, path, options.obj);
        out = this['__' + method](path, index, value, ver, options);
        refHelper.updateRefsForKey(path, ver, options);
        return out;
      };
    });
    adapter.__splice = adapter.splice;
    return adapter.splice = function() {
      var newMembers, options, out, path, removeCount, start, startIndex, ver, _ref2;
      _ref2 = argsNormalizer.adapter.splice(arguments), path = _ref2[0], start = _ref2[1], removeCount = _ref2[2], newMembers = _ref2[3], ver = _ref2[4], options = _ref2[5];
      options.obj || (options.obj = this._data);
      startIndex = refHelper.arrRefIndex(start, path, options.obj);
      out = this.__splice.apply(this, [path, startIndex, removeCount].concat(__slice.call(newMembers), [ver], [options]));
      refHelper.updateRefsForKey(path, ver, options);
      return out;
    };
  },
  arrRefIndex: function(start, path, obj) {
    var arr, i, mem, startIndex, _len;
    if ('number' === typeof start) {
      return start;
    }
    arr = this._adapter.lookup(path, true, {
      obj: obj
    }).obj;
    if (this.isArrayRef(path, obj)) {
      startIndex = arr.length;
      for (i = 0, _len = arr.length; i < _len; i++) {
        mem = arr[i];
        if (mem.id === start.id || parseInt(mem.id, 10) === parseInt(start.id, 10)) {
          return startIndex = i;
        }
      }
    }
    startIndex = arr.indexOf(start.id);
    if (startIndex !== -1) {
      return startIndex;
    }
    startIndex = arr.indexOf(parseInt(start.id, 10));
    if (startIndex !== -1) {
      return startIndex;
    }
    return arr.indexOf(start.id.toString());
  },
  ref: function(ref, key) {
    if (key != null) {
      return {
        $r: ref,
        $k: key
      };
    } else {
      return {
        $r: ref
      };
    }
  },
  arrayRef: function(ref, key) {
    return {
      $r: ref,
      $k: key,
      $t: 'array'
    };
  },
  $indexRefs: function(path, ref, key, type, ver, options) {
    var adapter, entry, keyOptions, keyVal, oldKey, oldRefObj, oldRefOptions, refHelper, refs, refsKey, refsKeys;
    adapter = this._adapter;
    refHelper = this;
    oldRefOptions = merge({
      dontFollowLastRef: true
    }, options);
    oldRefObj = adapter.lookup(path, false, oldRefOptions).obj;
    if (key) {
      entry = [ref, key];
      if (type) {
        entry.push(type);
      }
      adapter.lookup("$keys." + key + ".$", true, options).obj[path] = entry;
      keyVal = adapter.lookup(key, false, options).obj;
      if (type === void 0 && keyVal === void 0) {
        return;
      }
      if (type === 'array') {
        keyOptions = merge({
          array: true
        }, options);
        keyVal = adapter.lookup(key, true, keyOptions).obj;
        refsKeys = keyVal.map(function(keyValMem) {
          return ref + '.' + keyValMem;
        });
        this._removeOld$refs(oldRefObj, path, ver, options);
        return refsKeys.forEach(function(refsKey) {
          return refHelper._update$refs(refsKey, path, ref, key, type, options);
        });
      }
      refsKey = ref + '.' + keyVal;
    } else {
      if (oldRefObj && (oldKey = oldRefObj.$k)) {
        refs = adapter.lookup("$keys." + oldKey + ".$", false, options).obj;
        if (refs && refs[path]) {
          delete refs[path];
          if (!hasKeys(refs, {
            ignore: specHelper.reserved
          })) {
            adapter.del("$keys." + oldKey, ver, options);
          }
        }
      }
      refsKey = ref;
    }
    this._removeOld$refs(oldRefObj, path, ver, options);
    return this._update$refs(refsKey, path, ref, key, type, options);
  },
  _removeOld$refs: function(oldRefObj, path, ver, options) {
    var oldKey, oldKeyVal, oldRef, refHelper;
    if (oldRefObj && (oldRef = oldRefObj.$r)) {
      if (oldKey = oldRefObj.$k) {
        oldKeyVal = this._adapter.lookup(oldKey, false, options).obj;
      }
      if (oldKey && (oldRefObj.$t === 'array')) {
        refHelper = this;
        oldKeyVal.forEach(function(oldKeyMem) {
          return refHelper._removeFrom$refs(oldRef, oldKeyMem, path, ver, options);
        });
        return this._removeFrom$refs(oldRef, void 0, path, ver, options);
      } else {
        return this._removeFrom$refs(oldRef, oldKeyVal, path, ver, options);
      }
    }
  },
  _removeFrom$refs: function(ref, key, path, ver, options) {
    var refEntries, refWithKey;
    if (key) {
      refWithKey = ref + '.' + key;
    }
    refEntries = this._adapter.lookup("$refs." + refWithKey + ".$", false, options).obj;
    if (!refEntries) {
      return;
    }
    delete refEntries[path];
    if (!hasKeys(refEntries, {
      ignore: specHelper.reserved
    })) {
      return this._adapter.del("$refs." + ref, ver, options);
    }
  },
  _update$refs: function(refsKey, path, ref, key, type, options) {
    var entry, out;
    entry = [ref, key];
    if (type) {
      entry.push(type);
    }
    out = this._adapter.lookup("$refs." + refsKey + ".$", true, options);
    return out.obj[path] = entry;
  },
  updateRefsForKey: function(path, ver, options) {
    var refs;
    if (refs = this._adapter.lookup("$keys." + path + ".$", false, options).obj) {
      this._eachValidRef(refs, options.obj, __bind(function(path, ref, key, type) {
        return this.$indexRefs(path, ref, key, type, ver, options);
      }, this));
    }
    return this.eachValidRefPointingTo(path, options.obj, __bind(function(pointingPath, targetPathRemainder, ref, key, type) {
      return this.updateRefsForKey(pointingPath + '.' + targetPathRemainder, ver, options);
    }, this));
  },
  _eachValidRef: function(refs, obj, callback) {
    var fastLookup, key, o, path, ref, type, _ref2, _results;
    if (obj == null) {
      obj = this._adapter._data;
    }
    fastLookup = pathParser.fastLookup;
    _results = [];
    for (path in refs) {
      _ref2 = refs[path], ref = _ref2[0], key = _ref2[1], type = _ref2[2];
      if (path === '_proto') {
        continue;
      }
      o = fastLookup(path, obj);
      o = this._adapter.lookup(path, false, {
        obj: obj,
        dontFollowLastRef: true
      }).obj;
      _results.push(o && o.$r === ref && o.$k == key ? callback(path, ref, key, type) : delete refs[path]);
    }
    return _results;
  },
  _eachRefSetPointingTo: function(path, refs, fn) {
    var i, prop, props, refPos, refSet, _results;
    i = 0;
    refPos = refs;
    props = path.split('.');
    _results = [];
    while (prop = props[i++]) {
      if (!(refPos = refPos[prop])) {
        return;
      }
      _results.push((refSet = refPos.$) ? fn(refSet, props.slice(i).join('.'), prop) : void 0);
    }
    return _results;
  },
  eachValidRefPointingTo: function(targetPath, obj, fn) {
    var adapter, refs, self;
    self = this;
    adapter = self._adapter;
    if (!(refs = adapter.lookup('$refs', false, {
      obj: obj
    }).obj)) {
      return;
    }
    return self._eachRefSetPointingTo(targetPath, refs, function(refSet, targetPathRemainder, possibleIndex) {
      return self._eachValidRef(refSet, obj, function(pointingPath, ref, key, type) {
        if (type === 'array') {
          targetPathRemainder = possibleIndex + '.' + targetPathRemainder;
        }
        return fn(pointingPath, targetPathRemainder, ref, key, type);
      });
    });
  },
  eachArrayRefKeyedBy: function(path, obj, fn) {
    var key, ref, refSet, refs, type, _ref2, _results;
    if (!(refs = this._adapter.lookup("$keys", false, {
      obj: obj
    }).obj)) {
      return;
    }
    refSet = (path + '.$').split('.').reduce(function(refSet, prop) {
      return refSet && refSet[prop];
    }, refs);
    if (!refSet) {
      return;
    }
    _results = [];
    for (path in refSet) {
      _ref2 = refSet[path], ref = _ref2[0], key = _ref2[1], type = _ref2[2];
      _results.push(type === 'array' ? fn(path, ref, key) : void 0);
    }
    return _results;
  },
  notifyPointersTo: function(targetPath, obj, method, args, ignoreRoots) {
    if (ignoreRoots == null) {
      ignoreRoots = [];
    }
    this.eachValidRefPointingTo(targetPath, obj, __bind(function(pointingPath, targetPathRemainder, ref, key, type) {
      var id, index, rest, _ref2;
      if (type !== 'array') {
        if (this._alreadySeen(pointingPath, ref, ignoreRoots)) {
          return;
        }
        if (targetPathRemainder) {
          pointingPath += '.' + targetPathRemainder;
        }
      } else if (targetPathRemainder) {
        _ref2 = targetPathRemainder.split('.'), id = _ref2[0], rest = 2 <= _ref2.length ? __slice.call(_ref2, 1) : [];
        index = this._toIndex(key, id, obj);
        if (index !== -1) {
          pointingPath += '.' + index;
          if (rest.length) {
            pointingPath += '.' + rest.join('.');
          }
        }
      }
      return this._model.emit(method, [pointingPath].concat(__slice.call(args)));
    }, this));
    return this.eachArrayRefKeyedBy(targetPath, obj, __bind(function(pointingPath, ref, key) {
      var arrayMemberArgs, firstArgs, ns, _ref2;
      _ref2 = this.splitArrayArgs(method, args), firstArgs = _ref2[0], arrayMemberArgs = _ref2[1];
      if (arrayMemberArgs) {
        ns = this._adapter.lookup(ref, false, {
          obj: obj
        }).obj;
        arrayMemberArgs = arrayMemberArgs.map(function(arg) {
          return ns[arg] || ns[parseInt(arg, 10)];
        });
      }
      args = firstArgs.concat(arrayMemberArgs);
      return this._model.emit(method, [pointingPath].concat(__slice.call(args)));
    }, this));
  },
  _toIndex: function(arrayRefKey, id, obj) {
    var index, keyArr;
    keyArr = this._adapter.lookup(arrayRefKey, false, {
      array: true,
      obj: obj
    }).obj;
    index = keyArr.indexOf(id);
    if (index === -1) {
      return keyArr.indexOf(parseInt(id, 10));
    }
    return index;
  },
  _alreadySeen: function(pointingPath, ref, ignoreRoots) {
    var alreadySeen;
    alreadySeen = ignoreRoots.some(function(root) {
      return root === pointingPath.substr(0, root.length);
    });
    if (alreadySeen) {
      return true;
    }
    ignoreRoots.push(ref);
    return false;
  },
  cleanupPointersTo: function(path, options) {
    var adapter, key, keyMem, keyVal, model, pointingPath, ref, refs, _ref2, _results;
    adapter = this._adapter;
    refs = adapter.lookup("$refs." + path + ".$", false, options).obj;
    if (refs === void 0) {
      return;
    }
    model = this._model;
    _results = [];
    for (pointingPath in refs) {
      _ref2 = refs[pointingPath], ref = _ref2[0], key = _ref2[1];
      keyVal = key && adapter.lookup(key, false, options).obj;
      _results.push(keyVal && Array.isArray(keyVal) ? (keyMem = path.substr(ref.length + 1, pointingPath.length), adapter.remove(key, keyVal.indexOf(keyMem), 1, null, options)) : void 0);
    }
    return _results;
  },
  dereferenceTxn: function(txn, obj) {
    var $k, $r, args, method, newPushArgs, oldPushArgs, path, remainingProps, sliceFrom, _ref2, _ref3, _ref4;
    method = transaction.method(txn);
    args = transaction.args(txn);
    path = transaction.path(txn);
    if (ARRAY_OPS[method]) {
      sliceFrom = (function() {
        switch (method) {
          case 'push':
          case 'unshift':
            return 1;
          case 'pop':
          case 'shift':
          case 'insertAfter':
          case 'insertBefore':
            return 2;
          case 'remove':
          case 'move':
          case 'splice':
            return 3;
          default:
            throw new Error('Unimplemented for method ' + method);
        }
      })();
      if (_ref2 = this.isArrayRef(path, obj), $r = _ref2.$r, $k = _ref2.$k, _ref2) {
        this._model._cache.invalidateSpecModelCache();
        args[0] = path = $k;
        oldPushArgs = args.slice(sliceFrom);
        newPushArgs = oldPushArgs.map(function(refObjToAdd) {
          if (refObjToAdd.$r === void 0) {
            throw new Error('Trying to push a non-ref onto an array ref');
          }
          if ($r !== refObjToAdd.$r) {
            throw new Error("Trying to push elements of type " + refObjToAdd.$r + " onto path " + path + " that is an array ref of type " + $r);
          }
          return refObjToAdd.$k;
        });
        args.splice.apply(args, [sliceFrom, oldPushArgs.length].concat(__slice.call(newPushArgs)));
      } else {
        _ref3 = this._adapter.lookup(path, false, {
          obj: obj
        }), obj = _ref3.obj, path = _ref3.path, remainingProps = _ref3.remainingProps;
        if (obj === void 0 && (remainingProps != null ? remainingProps.length : void 0)) {
          args[0] = [path].concat(remainingProps).join('.');
        } else {
          args[0] = path;
        }
      }
      return txn;
    }
    _ref4 = this._adapter.lookup(path, false, {
      obj: obj
    }), obj = _ref4.obj, path = _ref4.path, remainingProps = _ref4.remainingProps;
    if (obj === void 0 && (remainingProps != null ? remainingProps.length : void 0)) {
      args[0] = [path].concat(remainingProps).join('.');
    } else {
      args[0] = path;
    }
    return txn;
  },
  isArrayRef: function(path, data) {
    var $k, $r, $t, options, refObj;
    options = {
      obj: data,
      dontFollowLastRef: true
    };
    refObj = this._adapter.lookup(path, false, options).obj;
    if (refObj === void 0) {
      return false;
    }
    $r = refObj.$r, $k = refObj.$k, $t = refObj.$t;
    if ($k) {
      $k = this.dereferencedPath($k, data);
    }
    if ($t === 'array') {
      return {
        $r: $r,
        $k: $k
      };
    } else {
      return false;
    }
  },
  splitArrayArgs: function(method, args) {
    switch (method) {
      case 'push':
        return [[], args];
      case 'insertAfter':
      case 'insertBefore':
        return [[args[0]], args.slice(1)];
      case 'splice':
        return [args.slice(0, 2), args.slice(2)];
      default:
        return [args, []];
    }
  },
  isRefObj: function(obj) {
    return '$r' in obj;
  },
  dereferencedPath: function(path, data) {
    var meta;
    meta = this._adapter.lookup(path, false, {
      obj: data,
      returnMeta: true
    });
    return meta.path;
  },
  denormalizePath: function(path, data) {
    var obj, remainingProps, _ref2;
    _ref2 = this._adapter.lookup(path, false, {
      obj: data
    }), path = _ref2.path, obj = _ref2.obj, remainingProps = _ref2.remainingProps;
    if (obj || !(remainingProps != null ? remainingProps.length : void 0)) {
      return path;
    }
    return [path].concat(remainingProps).join('.');
  }
};