var COMMIT, LOCK, LOCK_CLOCK_MASK, LOCK_TIMEOUT, LOCK_TIMEOUT_MASK, MAX_RETRIES, MAX_RETRY_DELAY, Stm, UNLOCK, transaction;
var __slice = Array.prototype.slice;
transaction = require('./transaction.server');
MAX_RETRIES = 10;
MAX_RETRY_DELAY = 10;
Stm = module.exports = function(redisClient) {
  var getLocks, lock;
  lock = function(len, locks, txnsSince, retries, callback) {
    return redisClient.eval.apply(redisClient, [LOCK, len].concat(__slice.call(locks), [txnsSince], [function(err, values) {
      if (err) {
        return callback(err);
      }
      if (values[0]) {
        return callback(null, values[0], values[1]);
      }
      if (retries) {
        return setTimeout(function() {
          return lock(len, locks, txnsSince, retries - 1, callback);
        }, Math.random() * MAX_RETRY_DELAY);
      }
      return callback('lockMaxRetries');
    }]));
  };
  this._getLocks = getLocks = function(path) {
    var lockPath, segment;
    lockPath = '';
    return ((function() {
      var _i, _len, _ref, _results;
      _ref = path.split('.');
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        segment = _ref[_i];
        _results.push(lockPath += '.' + segment);
      }
      return _results;
    })()).reverse();
  };
  this.commit = function(txn, callback) {
    var base, locks, locksLen, txnsSince;
    base = transaction.base(txn);
    txnsSince = base == null ? '' : base + 1;
    locks = getLocks(transaction.path(txn));
    locksLen = locks.length;
    return lock(locksLen, locks, txnsSince, MAX_RETRIES, function(err, lockVal, txns) {
      var conflict;
      if (err) {
        return callback(err);
      }
      if (txns && (conflict = transaction.journalConflict(txn, txns))) {
        return redisClient.eval.apply(redisClient, [UNLOCK, locksLen].concat(__slice.call(locks), [lockVal], [function(err) {
          if (err) {
            return callback(err);
          }
          return callback(conflict);
        }]));
      }
      return redisClient.eval.apply(redisClient, [COMMIT, locksLen].concat(__slice.call(locks), [lockVal], [JSON.stringify(txn)], [function(err, ver) {
        if (err) {
          return callback(err);
        }
        if (ver === 0) {
          return callback('lockReleased');
        }
        return callback(null, ver);
      }]));
    });
  };
};
Stm._LOCK_TIMEOUT = LOCK_TIMEOUT = 3;
Stm._LOCK_TIMEOUT_MASK = LOCK_TIMEOUT_MASK = 0x100000000;
Stm._LOCK_CLOCK_MASK = LOCK_CLOCK_MASK = 0x100000;
Stm._LOCK = LOCK = "local now = os.time()\nlocal path = KEYS[1]\nfor i, val in pairs(redis.call('smembers', path)) do\n  if val % " + LOCK_TIMEOUT_MASK + " < now then\n    redis.call('srem', path, val)\n  else\n    return 0\n  end\nend\nfor i, path in pairs(KEYS) do\n  path = 'l' .. path\n  local val = redis.call('get', path)\n  if val then\n    if val % " + LOCK_TIMEOUT_MASK + " < now then\n      redis.call('del', path)\n    else\n      return 0\n    end\n  end\nend\nlocal val = '0x' ..\n  string.format('%x', redis.call('incr', 'lockClock') % " + LOCK_CLOCK_MASK + ") ..\n  string.format('%x', now + " + LOCK_TIMEOUT + ")\nredis.call('set', 'l' .. path, val)\nfor i, path in pairs(KEYS) do\n  redis.call('sadd', path, val)\nend\nlocal txns\nif ARGV[1] ~= '' then txns = redis.call('zrangebyscore', 'txns', ARGV[1], '+inf') end\nreturn {val, txns}";
Stm._UNLOCK = UNLOCK = "local val = ARGV[1]\nlocal path = 'l' .. KEYS[1]\nif redis.call('get', path) == val then redis.call('del', path) end\nfor i, path in pairs(KEYS) do\n  redis.call('srem', path, val)\nend";
Stm._COMMIT = COMMIT = "local val = ARGV[1]\nlocal path = 'l' .. KEYS[1]\nlocal fail = false\nif redis.call('get', path) == val then redis.call('del', path) else fail = true end\nfor i, path in pairs(KEYS) do\n  if redis.call('srem', path, val) == 0 then return 0 end\nend\nif fail then return 0 end\nlocal ver = redis.call('incr', 'ver')\nredis.call('zadd', 'txns', ver, ARGV[2])\nreturn ver";