var Memory, merge, specHelper;
var __slice = Array.prototype.slice, __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
merge = require('../util').merge;
specHelper = require('../specHelper');
Memory = module.exports = function() {
  this._data = {};
  this.ver = 0;
};
Memory.prototype = {
  get: function(path, obj) {
    if (obj == null) {
      obj = this._data;
    }
    if (path) {
      return this.lookup(path, false, {
        obj: obj
      }).obj;
    } else {
      return obj;
    }
  },
  set: function(path, value, ver, options) {
    var obj, out, parent, prop, refObjCopy, _ref;
    if (options == null) {
      options = {};
    }
    if (value && value.$r) {
      refObjCopy = merge({}, value);
      value = refObjCopy;
    }
    this.ver = ver;
    _ref = out = this.lookup(path, true, options), parent = _ref.parent, prop = _ref.prop;
    obj = out.obj = parent[prop] = value;
    if (options.returnMeta) {
      return out;
    } else {
      return obj;
    }
  },
  del: function(path, ver, options) {
    var curr, key, obj, out, parent, parentProto, prop, value, _ref;
    if (options == null) {
      options = {};
    }
    this.ver = ver;
    _ref = out = this.lookup(path, false, options), parent = _ref.parent, prop = _ref.prop, obj = _ref.obj, path = _ref.path;
    if (!parent) {
      if (options.returnMeta) {
        return out;
      } else {
        return obj;
      }
    }
    if (options.proto) {
      parentProto = Object.getPrototypeOf(parent);
      if (prop in parentProto) {
        curr = {};
        for (key in parentProto) {
          value = parentProto[key];
          if (key !== prop) {
            curr[key] = typeof value === 'object' ? Object.create(value) : value;
          }
        }
        parent.__proto__ = curr;
      }
    }
    delete parent[prop];
    if (options.returnMeta) {
      return out;
    } else {
      return obj;
    }
  },
  push: function() {
    var arr, options, out, path, ret, values, ver, _i;
    path = arguments[0], values = 4 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 2) : (_i = 1, []), ver = arguments[_i++], options = arguments[_i++];
    if (options === void 0) {
      options = {};
    }
    if (options.constructor !== Object) {
      values.push(ver);
      ver = options;
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    ret = arr.push.apply(arr, values);
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  pop: function(path, ver, options) {
    var arr, out, ret;
    if (options == null) {
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    ret = arr.pop();
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  insertAfter: function(path, afterIndex, value, ver, options) {
    var arr, out, ret;
    if (options == null) {
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    if (!((-1 <= afterIndex && afterIndex <= arr.length - 1))) {
      throw new Error('Out of Bounds');
    }
    ret = arr.splice(afterIndex + 1, 0, value);
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  insertBefore: function(path, beforeIndex, value, ver, options) {
    var arr, out, ret;
    if (options == null) {
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    if (!((0 <= beforeIndex && beforeIndex <= arr.length))) {
      throw new Error('Out of Bounds');
    }
    ret = arr.splice(beforeIndex, 0, value);
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  remove: function(path, startIndex, howMany, ver, options) {
    var arr, out, ret, upperBound;
    if (options == null) {
      options = {};
    }
    this.ver = ver;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    upperBound = arr.length ? arr.length - 1 : 0;
    if (!((0 <= startIndex && startIndex <= upperBound))) {
      throw new Error('Out of Bounds');
    }
    ret = arr.splice(startIndex, howMany);
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  splice: function() {
    var arr, newMembers, options, out, path, removeCount, ret, startIndex, ver, _i;
    path = arguments[0], startIndex = arguments[1], removeCount = arguments[2], newMembers = 6 <= arguments.length ? __slice.call(arguments, 3, _i = arguments.length - 2) : (_i = 3, []), ver = arguments[_i++], options = arguments[_i++];
    if (options === void 0) {
      options = {};
    }
    if (options.constructor !== Object) {
      newMembers.push(ver);
      ver = options;
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    ret = arr.splice.apply(arr, [startIndex, removeCount].concat(__slice.call(newMembers)));
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  unshift: function() {
    var arr, newMembers, options, out, path, ret, ver, _i;
    path = arguments[0], newMembers = 4 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 2) : (_i = 1, []), ver = arguments[_i++], options = arguments[_i++];
    if (options == null) {
      options = {};
    }
    if (options === void 0) {
      options = {};
    }
    if (options.constructor !== Object) {
      newMembers.push(ver);
      ver = options;
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    ret = arr.unshift.apply(arr, newMembers);
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  shift: function(path, ver, options) {
    var arr, out, ret;
    if (options == null) {
      options = {};
    }
    this.ver = ver;
    options.array = true;
    out = this.lookup(path, true, options);
    arr = out.obj;
    if (!specHelper.isArray(arr)) {
      throw new Error('Not an Array');
    }
    ret = arr.shift();
    if (options.returnMeta) {
      return out;
    } else {
      return ret;
    }
  },
  move: function(path, from, to, ver, options) {
    var value;
    if (options == null) {
      options = {};
    }
    value = this.lookup("" + path + "." + from, false, options).obj;
    if (to < 0) {
      to += this.lookup(path, false, options).obj.length;
    }
    if (from > to) {
      this.insertBefore(path, to, value, ver, options);
      from++;
    } else {
      this.insertAfter(path, to, value, ver, options);
    }
    return this.remove(path, from, 1, ver, options);
  },
  lookup: function(path, addPath, options) {
    var arrIndex, array, i, key, keyVal, len, next, parent, pathSegment, prop, props, proto, ref;
    proto = options.proto, array = options.array;
    next = options.obj || this._data;
    props = path.split('.');
    path = '';
    i = 0;
    len = props.length;
    while (i < len) {
      parent = next;
      prop = props[i++];
      if (proto && !specHelper.isSpeculative(parent)) {
        parent = specHelper.create(parent);
      }
      pathSegment = prop.dereffedProp || prop;
      prop = prop.arrIndex !== void 0 ? prop.arrIndex : prop;
      path = path ? path + '.' + pathSegment : pathSegment;
      next = parent[prop];
      if (next === void 0) {
        if (!addPath) {
          return {
            obj: next,
            path: path,
            remainingProps: props.slice(i)
          };
        }
        if (array && i === len) {
          next = proto ? specHelper.create([]) : [];
        } else {
          next = proto ? specHelper.create({}) : {};
        }
        parent[prop] = next;
      } else if (proto && typeof next === 'object' && !specHelper.isSpeculative(next)) {
        if (specHelper.isSpeculative(parent) && parent.hasOwnProperty(prop)) {
          next._proto = true;
        } else {
          next = specHelper.create(next);
        }
        parent[prop] = next;
      }
      if (ref = next.$r) {
        if (i === len && options.dontFollowLastRef) {
          return {
            path: path,
            parent: parent,
            prop: prop,
            obj: next
          };
        }
        if (!(key = next.$k)) {
          next = this.lookup(ref, addPath, options).obj;
        } else {
          keyVal = this.lookup(key, false, options).obj;
          if (specHelper.isArray(keyVal)) {
            next = keyVal.map(__bind(function(key) {
              return this.lookup(ref + '.' + key, false, options).obj;
            }, this));
            if (props[i]) {
              props[i] = {
                arrIndex: arrIndex = parseInt(props[i], 10),
                dereffedProp: keyVal[arrIndex]
              };
            }
          } else {
            ref = ref + '.' + this.lookup(key, false, options).obj;
            next = this.lookup(ref, addPath, options).obj;
          }
        }
        if (i < len) {
          path = ref;
        }
        if (next === void 0 && !addPath && i < len) {
          return {
            obj: next,
            path: path,
            remainingProps: props.slice(i)
          };
        }
      }
    }
    return {
      path: path,
      parent: parent,
      prop: prop,
      obj: next
    };
  }
};