var Subscription, events;
var __slice = Array.prototype.slice;
Subscription = function(target, event, fn, context) {
  this.target = target;
  this.event = event;
  this.fn = fn;
  this.context = context;
  this.active = true;
};
Subscription.prototype.cancel = function() {
  if (!this.active) {
    return;
  }
  this.target.removeListener(this.event, this.fn, this.context);
  return this.active = false;
};
module.exports = events = {
  Subscription: Subscription,
  on: function(ev, fn, context) {
    var cbs, ev2cb;
    ev2cb = this._listeners || (this._listeners = {});
    cbs = ev2cb[ev] || (ev2cb[ev] = []);
    cbs.push([fn, context]);
    return new Subscription(this, ev, fn, context);
  },
  removeListener: function(ev, fn, context) {
    var cb, cbs, ev2cb, i;
    if (!(ev2cb = this._listeners)) {
      return this;
    }
    if (!(cbs = ev2cb[ev])) {
      return this;
    }
    i = cbs.length;
    while (i--) {
      cb = cbs[i];
      if (cb[0] === fn && cb[1] === context) {
        cbs.splice(i, 1);
      }
    }
    return this;
  },
  trigger: function() {
    var args, cb, cbs, ev, ev2cb, i, lastArg, mute;
    ev = arguments[0], args = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
    ev2cb = this._listeners;
    lastArg = args[args.length - 1];
    mute = null;
    if (lastArg && typeof lastArg === 'object') {
      mute = lastArg.mute;
    }
    if (!ev2cb) {
      return this;
    }
    if (!(cbs = ev2cb[ev])) {
      return this;
    }
    i = cbs.length;
    while (i--) {
      cb = cbs[i];
      if (mute && this._shouldMute(mute, cb)) {
        continue;
      }
      cb[0].apply(cb[1], args);
    }
    return this;
  },
  _shouldMute: function(mute, cb) {
    var j;
    if (j = mute.length) {
      while (j--) {
        if (this._shouldMute(mute[j], cb)) {
          return true;
        }
      }
    } else if (mute.target === this && (!(mute.context && cb[1]) || mute.context === cb[1]) && mute.fn === cb[0]) {
      return true;
    }
    return false;
  },
  once: function(ev, fn, context) {
    var subsc;
    subsc = this.on(ev, function() {
      subsc.cancel();
      return fn.apply(this, Array.prototype.slice.call(arguments));
    }, context);
    return subsc;
  }
};